package results;

import java.io.File;
import java.io.IOException;
import java.text.Collator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import java.util.TreeSet;
import java.util.Vector;

/** Contest */
public final class Contest
{
  /** Number of privileged jurors in the jury */
  private static final int numPrivilegedJurors = 1;
  /** Series of the contest */
  Series series;
  /** Name of the contest */
  String name;
  /** Jurors for this competition */
  Vector judges;
  /** Participants in this competititon */
  Vector mirrors;
  /** Amounts of different mirror numbers (for duplicate detection) */
  private transient Hashtable numbers = new Hashtable ();
  /** Picture format (null=no format) */
  private String format;

  /** Flag the data modified */
  void flagModified () {
    series.flagModified ();
    for (Enumeration en = mirrors.elements (); en.hasMoreElements (); )
      ((Mirror) en.nextElement ()).flagModifiedContest ();
  }

  /** Constructor
   * @param name	Name of the contest
   */
  public Contest (String name) {
    this (name, new Vector ());
  }

  /** Constructor
   * @param name	Name of the contest
   * @param judges	The jury of this contest
   */
  Contest (String name, Vector judges) {
    this.name = name;
    this.judges = judges;
    this.mirrors = new Vector ();
  }

  /** Convert the contest to a printable string */
  public String toString () { return name; }

  /** Get the name of the contest */
  public String getName () { return name; }
  /** Set the name of the contest */
  public void setName (String name) {
    if (!this.name.equals (name)) {
      this.name = name;
      if (series != null) series.flagModified ();
    }
  }
  /** Get the picture format of the contest */
  public String getFormat () { return format; }
  /** Set the picture format of the contest */
  public void setFormat (String format) {
    if (format != null && format.equals (""))
      format = null;
    if (this.format == format)
      return;
    if (this.format != null && format != null &&
	this.format.equals (format))
      return;
    this.format = format;
    if (series != null) series.flagModified ();
  }

  /** Get the number of registered competitors */
  public int getNumMirrors () { return mirrors.size (); }
  /** Get the nth mirror */
  public Mirror getMirror (int i) { return (Mirror) mirrors.elementAt (i); }
  /** Get the mirror of a competitor
   * @param competitor	the competitor
   * @return		the mirror, or null if none registered
   */
  public Mirror getMirror (Competitor competitor) {
    for (Enumeration en = mirrors.elements (); en.hasMoreElements (); ) {
      Mirror mirror = (Mirror) en.nextElement ();
      if (mirror.competitor == competitor)
	return mirror;
    }
    return null;
  }
  /** Get the number of occurrences of a specific mirror number */
  public int getNumOccurrences (Integer number) {
    if (number == null || number.intValue () == 0) return 0;
    Integer count = (Integer) numbers.get (number);
    return count == null ? 0 : count.intValue ();
  }
  /** Increment the number of occurrences of a specific mirror number */
  final void incNumbers (int number) {
    if (number == 0) return;
    Integer num = new Integer (number);
    Integer count = (Integer) numbers.get (num);
    numbers.put (num, new Integer (count == null
				   ? 1
				   : count.intValue () + 1));
  }
  /** Decrement the number of occurrences of a specific mirror number */
  final void decNumbers (int number) {
    if (number == 0) return;
    Integer num = new Integer (number);
    numbers.put (num, new Integer
      (((Integer) numbers.get (num)).intValue () - 1));
  }

  /** Get the number of registered judges */
  public int getNumJudges () { return judges.size (); }
  /** Get the nth judge */
  public Judge getJudge (int i) { return (Judge) judges.elementAt (i); }
  /** Swap two judges at specified positions */
  public void swapJudges (int i, int j) {
    Judge judge = (Judge) judges.elementAt (i);
    judges.setElementAt (judges.elementAt (j), i);
    judges.setElementAt (judge, j);
    flagModified ();
  }
  /** Determine if a judge belongs to the contest */
  public boolean hasJudge (Judge judge) { return judges.contains (judge); }

  /** Register a competitor to the contest
   * @param competitor	Competitor to be registered
   */
  public void add (Competitor competitor) {
    add (new Mirror (competitor, this, 0));
  }
  /** Add a mirror to the contest
   * @param mirror	mirror to be added
   */
  void add (Mirror mirror) {
    if (mirrors.contains (mirror))
      return;
    mirrors.addElement (mirror);
    incNumbers (mirror.number);
    flagModified ();
  }
  /** Register a juror to the contest
   * @param judge	the juror
   */
  public void add (Judge judge) {
    judges.addElement (judge);
    flagModified ();
  }
  /** Remove a judge from the contest
   * @param judge	the judge to be removed
   * @return		true if the judge belonged to the contest
   */
  public boolean remove (Judge judge)
  {
    boolean removed = judges.remove (judge);
    if (removed) flagModified ();
    return removed;
  }
  /** Remove a mirror from the contest
   * @param mirror	the mirror to be removed
   * @return		true if the mirror belonged to the contest
   */
  public boolean remove (Mirror mirror)
  {
    if (!mirrors.remove (mirror))
      return false;
    decNumbers (mirror.number);
    flagModified ();
    return true;
  }
  /** Remove a competitor from the contest
   * @param competitor	the competitor to be removed
   * @return		true if the competitor belonged to the contest
   */
  public boolean remove (Competitor competitor)
  {
    return remove (getMirror (competitor));
  }
  /** Remove all competitors from the contest */
  public void removeCompetitors ()
  {
    if (!mirrors.isEmpty ()) {
      mirrors = new Vector ();
      numbers = new Hashtable ();
      flagModified ();
    }
  }
  /** Add all competitors from the series */
  public void addCompetitors ()
  {
    int i = series.getNumCompetitors ();
    if (i == mirrors.size ())
      return;
    while (i-- > 0)
      add (series.getCompetitor (i));
  }

  /** Shuffle the mirror numbers
   * @param first	lowest mirror number
   */
  public void shuffleMirrors (int first)
  {
    int i;
    Mirror[] nums = new Mirror[mirrors.size ()];
    mirrors.copyInto (nums);
    numbers.clear ();
    for (i = 0; i < nums.length; i++)
      numbers.put (new Integer (first + i), new Integer (1));
    Random random = new Random ();
    while (i > 0) {
      int index = Math.abs (random.nextInt ()) % i;
      Mirror m = nums[--i];
      (nums[i] = nums[index]).number = first + i;
      nums[index] = m;
    }
    for (i = nums.length; i-- > 0; )
      mirrors.setElementAt (nums[i], i);
    flagModified ();
  }

  /** Get an bitmap of points given too frequently by a judge */
  public boolean[] getFrequentPoints (Judge judge)
  {
    int i = mirrors.size ();
    int points[] = new int[i];
    while (i-- > 0) {
      Integer p = ((Mirror) mirrors.elementAt (i)).getPoints (judge);
      points[i] = p == null ? 0 : p.intValue ();
    }
    return series.constraints.getFrequent (points);
  }
  /** Get an bitmap of points given too infrequently by a judge */
  public boolean[] getInfrequentPoints (Judge judge)
  {
    int i = mirrors.size ();
    int points[] = new int[i];
    while (i-- > 0) {
      Integer p = ((Mirror) mirrors.elementAt (i)).getPoints (judge);
      points[i] = p == null ? 0 : p.intValue ();
    }
    return series.constraints.getInfrequent (points);
  }
  /** Determine if a judge has given any points */
  public boolean hasPoints (Judge judge)
  {
    for (int i = mirrors.size (); i-- > 0; )
      if (((Mirror) mirrors.elementAt (i)).getPoints (judge) != null)
	return true;
    return false;
  }

  /** Rectify the points given by the jurors */
  public void rectify () {
    for (Enumeration m = mirrors.elements (); m.hasMoreElements (); )
      ((Mirror) m.nextElement ()).rectify ();
  }

  /** Generate a list of the mirror numbers in the competition
   * @param dir		the directory where the list is to be created
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateMirrors (File dir, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "mirrors.html", series.locale);
    html.header (series.name + " - " + name, "mirrors.index");
    boolean printedLine = false;
    TreeSet sortedMirrors = new TreeSet (new MirrorSorter (reporter));
    sortedMirrors.addAll (mirrors);
    for (Iterator it = sortedMirrors.iterator (); it.hasNext (); ) {
      Mirror mirror = (Mirror) it.next ();
      if (mirror.number == 0) {
	reporter.reportMissingNumber (mirror);
	continue;
      }
      if (!printedLine) {
	printedLine = true;
	html.out.println ("<TABLE CELLPADDING=10 CELLSPACING=5>\n");
      }
      html.out.println ("<TR>");
      html.out.print ("<TD>");
      html.out.print (mirror.number);
      html.out.println ("</TD>");
      html.out.print ("<TD>");
      html.html (mirror.competitor.name);
      html.out.println ("</TD>");
      html.out.print ("<TD>");
      html.html (mirror.competitor.country.name);
      html.out.println ("</TD>");
      html.out.print ("<TD>");
      if (mirror.competitor.team != null)
	html.html (mirror.competitor.team.name);
      html.out.println ("</TD>");
      html.out.println ("</TR>");
    }
    if (printedLine)
      html.out.println ("</TABLE>\n");
    html.footer ();
  }

  /** Generate individual results for the contest
   * @param dir		the directory where the file is to be created
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateResults (File dir, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "results.html", series.locale);
    html.header (series.name + " - " + name, "results.index");

    html.out.print ("<H3>");
    html.link ("results.judges", "judges.html", null, null);
    html.out.println ("</H3>\n");

    int jurysize = 0;
    for (Enumeration j = judges.elements (); j.hasMoreElements (); ) {
      Judge judge = (Judge) j.nextElement ();
      if (jurysize == 0)
	html.out.print ("<OL>\n");
      html.out.print ("<LI>");
      html.format ("results.judge", new Object[]
	{ judge.name, judge.country.name,
	  jurysize++ == 0 ? html.lookup ("results.president") : "" });
      html.out.println ("</LI>");
    }
    if (jurysize > 0)
      html.out.println ("</OL>\n");

    html.out.println ("<TABLE BORDER>");
    html.out.println ("<TR>");
    html.tag ("TD", "results.rank");
    html.tag ("TD", "results.mirror");
    html.tag ("TD", "results.name");
    html.tag ("TD", "results.team");
    for (int i = 1; i <= jurysize; i++)
      html.out.println ("<TD>" + i + "</TD>");
    html.tag ("TD", "results.total");
    html.tag ("TD", "results.penalty");
    html.tag ("TD", "results.final");
    html.out.println ("</TR>");

    // flags: which "cards" have been issued
    boolean red = false, yellow = false, grey = false;
    // penalty reasons that have occurred
    TreeSet penalties = new TreeSet (new ReasonSorter ());
    // penalty reason explanations
    Reasons reasons = new Reasons (series.locale);

    // Display the rankings
    for (Enumeration e = getRanking (); e.hasMoreElements (); ) {
      Mirror mirror = (Mirror) e.nextElement ();
      html.out.print ("<TR>\n<TD ALIGN=RIGHT>" + mirror.rank +
		      "</TD>\n<TD ALIGN=RIGHT><A NAME=\"" + mirror.number +
		      "\">" + mirror.number + "</A>");
      if (format != null)
	html.img ("", new Integer (mirror.number).toString (), format);
      html.out.print ("</TD>\n<TD>");
      html.html (mirror.competitor.name);
      html.out.print ("</TD>\n<TD>");
      html.html (mirror.competitor.team == null
		 ? mirror.competitor.country.name
		 : mirror.competitor.team.name + ", " +
		 mirror.competitor.country.name);
      html.out.print ("</TD>");
      for (Enumeration j = judges.elements (); j.hasMoreElements (); ) {
	Judge judge = (Judge) j.nextElement ();
	Points points = mirror.getPts (judge);
	if (points == null || points.orig == null) {
	  html.out.print ("<TD ALIGN=CENTER>-</TD>");
	  reporter.reportMissingPoints (mirror, judge);
	}
	else {
	  int p = points.orig.intValue ();
	  int average = mirror.rectify ();
	  int rectified = points.rect.intValue ();
	  if (p - average >= 5 || average - p >= 5) {
	    html.out.print ("<TD ALIGN=RIGHT BGCOLOR=\"#ff5f5f\">");
	    red = true;
	  }
	  else if (!series.national &&
		   judge.country == mirror.competitor.country) {
	    html.out.print ("<TD ALIGN=RIGHT BGCOLOR=\"#808080\">");
	    grey = true;
	  }
	  else if (p != rectified) {
	    html.out.print ("<TD ALIGN=RIGHT BGCOLOR=\"#ffffbf\">");
	    yellow = true;
	  }
	  else
	    html.out.print ("<TD ALIGN=RIGHT>");
	  html.out.print (rectified);
	  if (p != rectified)
	    html.out.print ("<BR><SMALL>" + p + "</SMALL>");
	  html.out.print ("</TD>");
	}
      }

      html.out.println ("<TD ALIGN=RIGHT>" +
			mirror.getTotalRectifiedNoPenalty () +
			"<BR><SMALL>" + mirror.getTotalNoPenalty () +
			"</SMALL></TD>\n<TD ALIGN=CENTER>" +
			(mirror.penalty != 0
			 ? (mirror.penalty + " (" + mirror.penaltyReason + ")")
			 : "-") +
			"</TD>\n<TD ALIGN=RIGHT>" +
			mirror.getTotalRectified () + "</TD>\n</TR>");
      if (mirror.penalty != 0)
	penalties.add (reasons.get
		       ((new Character (mirror.penaltyReason)).toString ()));
    }

    html.out.println ("</TABLE>\n");

    if (penalties.size () > 0) {
      html.out.println ("<TABLE>");
      for (Iterator it = penalties.iterator (); it.hasNext (); ) {
	Reason reason = (Reason) it.next ();
	html.out.print ("<TR><TD>" + reason.code + "</TD>\n<TD>");
	html.html (reason.toString ());
	html.out.println ("</TD></TR>");
      }
      html.out.println ("</TABLE>\n");
    }

    if (red || yellow || grey) {
      html.out.println ("<TABLE>");
      if (red) {
	html.out.println ("<TR>\n<TD BGCOLOR=\"#ff5f5f\"></TD>");
	html.tag ("TD", "results.red");
	html.tag ("TD", "results.red.d");
	html.out.println ("</TR>");
      }
      if (yellow) {
	html.out.println ("<TR>\n<TD BGCOLOR=\"#ffffbf\"></TD>");
	html.tag ("TD", "results.yellow");
	html.tag ("TD", "results.yellow.d");
	html.out.println ("</TR>");
      }
      if (grey) {
	html.out.println ("<TR>\n<TD BGCOLOR=\"#808080\"></TD>");
	html.tag ("TD", "results.grey");
	html.tag ("TD", "results.grey.d");
	html.out.println ("</TR>");
      }
      html.out.println ("</TABLE>");
    }
    html.footer ();
    generateJurors (dir, reporter);
  }

  /** Generate award list for the contest
   * @param html	the HTML output stream
   * @param top		number of top ranks to list
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateAwards (HTMLWriter html, int top, ErrorReporter reporter)
    throws IOException
  {
    // Display the rankings
    for (Enumeration e = getRanking (); e.hasMoreElements (); ) {
      final Mirror mirror = (Mirror) e.nextElement ();
      if (mirror.rank > top)
	break;
      html.out.println ("<P>" + mirror.rank + "." + mirror.competitor.name +
			";" + (mirror.competitor.team == null
			       ? mirror.competitor.country.name
			       : mirror.competitor.team.name + ", " +
			       mirror.competitor.country.name) + "("
			+ mirror.getTotalRectified () + ")</P>");
    }
  }


  /** Generate juror statistics for the contest
   * @param dir		the directory where the file is to be created
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateJurors (File dir, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "judges.html", series.locale);
    html.header (series.name + " - " + name, "judges.index");

    /** Points given indexed by juror */
    Hashtable jurors = new Hashtable ();
    /** Errors indexed by juror */
    Hashtable errors = new Hashtable ();
    /** All different points given by different judges */
    Vector givenPoints = new Vector ();

    // Gather the statistics
    for (Enumeration j = judges.elements (); j.hasMoreElements (); ) {
      Judge judge = (Judge) j.nextElement ();
      /** Mirrors indexed by points given */
      Hashtable points = new Hashtable ();
      /** Number of red cards */
      int numRed = 0;
      /** Number of yellow cards (rectifications), not counting "grey cards" */
      int numYellow = 0;
      /** Total effect of all rectifications */
      int totalRectified = 0;
      for (Enumeration m = mirrors.elements (); m.hasMoreElements (); ) {
	Mirror mirror = (Mirror) m.nextElement ();
	Points p = mirror.getPts (judge);
	if (p != null && p.orig != null) {
	  /* Gather rectification statistics */
	  int average = mirror.computeRectified ();
	  int rectified = p.rect.intValue ();
	  int diff = Math.abs (p.orig.intValue () - average);
	  if (diff >= 5)
	    numRed++;
	  else if (p.orig.intValue () != rectified &&
		   (series.national ||
		    judge.country != mirror.competitor.country))
	    numYellow++;
	  totalRectified += diff;
	  if (!givenPoints.contains (p.orig)) {
	    int i;
	    for (i = 0; i < givenPoints.size (); i++) {
	      if (((Integer) givenPoints.elementAt (i)).intValue () <
		  p.orig.intValue ()) {
		givenPoints.insertElementAt (p.orig, i);
		break;
	      }
	    }
	    if (i == givenPoints.size ())
	      givenPoints.addElement (p.orig);
	  }
	  Vector pointmirrors = (Vector) points.get (p.orig);
	  if (pointmirrors == null)
	    points.put (p.orig, pointmirrors = new Vector ());
	  pointmirrors.addElement (mirror);
	}
      }
      jurors.put (judge, points);
      errors.put (judge, new int[] {numRed, numYellow, totalRectified});
    }

    // Display the statistics
    html.out.println ("<TABLE BORDER=2>\n<TR>");
    html.tag ("TD", "judges.name");
    html.tag ("TD", "judges.red");
    html.tag ("TD", "judges.yellow");
    html.tag ("TD", "judges.total");
    for (Enumeration p = givenPoints.elements (); p.hasMoreElements (); )
      html.format ("judges.mirror", new Object[] { p.nextElement () });
    html.out.println ("</TR>\n");

    for (Enumeration j = judges.elements (); j.hasMoreElements (); ) {
      Judge judge = (Judge) j.nextElement ();
      Hashtable points = (Hashtable) jurors.get (judge);
      int err[] = (int[]) errors.get (judge);

      html.out.println ("<TR>\n<TD>" + judge.name + "</TD>\n<TD>" +
			err[0] + "</TD>\n<TD>" +
			err[1] + "</TD>\n<TD>" +
			err[2] + "</TD>");

      for (Enumeration p = givenPoints.elements (); p.hasMoreElements (); ) {
	Vector pmirrors = (Vector) points.get (p.nextElement ());
	if (pmirrors != null) {
	  html.out.print ("<TD>");
	  for (Enumeration m = pmirrors.elements (); m.hasMoreElements (); ) {
	    Mirror mirror = (Mirror) m.nextElement ();
	    html.out.print (mirror.rank);
	    if (m.hasMoreElements ())
	      html.out.print (",");
	  }
	  html.out.print ("<BR>");
	  for (Enumeration m = pmirrors.elements (); m.hasMoreElements (); ) {
	    Mirror mirror = (Mirror) m.nextElement ();
	    html.link (mirror.number, "results.html",
		       null, Integer.toString (mirror.number));
	    if (m.hasMoreElements ())
	      html.out.print (",");
	  }
	  html.out.println ("</TD>");
	}
	else
	  html.out.println ("<TD>-</TD>");
      }
      html.out.println ("</TR>");

      // Complain about frequent and infrequent points
      boolean[] freq = getFrequentPoints (judge);
      for (int i = 0; i < freq.length; i++)
	if (freq[i])
	  reporter.reportFrequent (i, judge, this);
      freq = getInfrequentPoints (judge);
      for (int i = 0; i < freq.length; i++)
	if (freq[i])
	  reporter.reportInfrequent (i, judge, this);
    }
    html.out.println ("</TABLE>");
    html.footer ();
  }

  /** Determine the rankings
   * @return	Enumeration to a vector of mirrors in the ranking order
   */
  Enumeration getRanking ()
  {
    Vector ranking = new Vector ();
    Collator collator = Collator.getInstance (series.locale);

    for (Enumeration m = mirrors.elements (); m.hasMoreElements (); ) {
      Mirror mirror1 = (Mirror) m.nextElement ();
      int i;
    cmp:
      for (i = 0; i < ranking.size (); i++) {
	Mirror mirror2 = (Mirror) ranking.elementAt (i);
	int points1 = mirror1.getTotalRectified ();
	int points2 = mirror2.getTotalRectified ();
	if (points1 > points2)
	  break;
	if (points1 < points2)
	  continue;
	if (judges.size () > 0) {
	  Integer p1[] = mirror1.getPoints ();
	  Integer p2[] = mirror2.getPoints ();
	  for (int j = 0; j < numPrivilegedJurors; j++) {
	    if (p1[j] == null || p2[j] == null)
	      continue;
	    points1 = p1[j].intValue () * p1.length - mirror1.penalty;
	    points2 = p2[j].intValue () * p2.length - mirror2.penalty;
	    if (points1 > points2)
	      break cmp;
	    if (points1 < points2)
	      continue cmp;
	  }
	}

	// Still even game: compare number of highest points
	PointJudges pj1[] = mirror1.getHistogram ();
	PointJudges pj2[] = mirror2.getHistogram ();

	for (int j = 0; j < pj1.length && j < pj2.length; j++) {
	  points1 = pj1[j].points;
	  points2 = pj2[j].points;
	  if (points1 < points2)
	    break cmp;
	  if (points1 > points2)
	    continue cmp;
	  points1 = pj1[j].getNumJudges ();
	  points2 = pj2[j].getNumJudges ();
	  if (points1 < points2)
	    break cmp;
	  if (points1 > points2)
	    continue cmp;
	}

	// It is an even game: collate the competitors in alphabetical order
	if (collator.compare (mirror1.competitor.name,
			      mirror2.competitor.name) < 0) {
	  mirror2.evengame = true;
	  break;
	}
	else
	  mirror1.evengame = true;
      }
      if (i == ranking.size ())
	ranking.addElement (mirror1);
      else
	ranking.insertElementAt (mirror1, i);
    }

    for (int place = 0, rank = 0; place < ranking.size (); place++) {
      Mirror mirror = (Mirror) ranking.elementAt (place);
      if (!mirror.evengame)
	rank = place + 1;
      mirror.rank = rank;
    }
    return ranking.elements ();
  }
}
