package results;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Locale;
import java.util.ResourceBundle;
import java.text.MessageFormat;

/** HTML utility class */
public final class HTMLWriter
{
  /** The resource bundle for localized strings */
  private ResourceBundle bundle;
  /** The message format for formatted output */
  private MessageFormat mf;
  /** The output stream */
  PrintWriter out;

  /** Constructor
   * @param dir		the output directory
   * @param name	the file name label
   * @param locale	the locale
   * @exception		IOException if the output file cannot be created
   */
  public HTMLWriter (File dir, String name, Locale locale)
    throws IOException
  {
    bundle = ResourceBundle.getBundle ("html", locale);
    (mf = new MessageFormat ("")).setLocale (locale);
    out = new PrintWriter (new FileWriter
      (new File (dir, bundle.getString (name))));
  }

  /** Called by the garbage collector before the object is discarded */
  protected void finalize ()
  {
    if (out != null)
      footer ();
  }

  /** Generate a translation of a label
   * @param label	the label
   */
  public void translate (String label)
  {
    out.print (bundle.getString (label));
  }

  /** Format a translation of a label
   * @param label	the label
   * @param params	the parameters to format
   */
  public void format (String label, Object[] params)
  {
    mf.applyPattern (bundle.getString (label));
    out.print (mf.format (params));
  }

  /** Look up a translation of a label
   * @param label	the label
   * @return		the translation
   */
  public String lookup (String label)
  {
    return bundle.getString (label);
  }

  /** Generate an HTML header
   * @param title	Main title
   * @param subtitle	Subtitle (translated if non-null)
   */
  public void header (String title, String subtitle)
  {
    out.print ("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n\n<HTML><HEAD>\n<TITLE>");
    html (title);
    out.print ("</TITLE>\n</HEAD>\n<BODY BGCOLOR=\"#ffffff\" LINK=\"#00007f\" ALINK=\"#0000ff\" VLINK=\"#00003f\">\n\n<H1>");
    html (title);
    out.println ("</H1>");

    if (subtitle != null) {
      out.print ("<H2>");
      translate (subtitle);
      out.println ("</H2>");
    }
  }

  /** Generate an HTML footer */
  public void footer ()
  {
    out.println ("</BODY></HTML>");
    out.close ();
    out = null;
  }

  /** Generate a HTML tag with translated content
   * @param tag		the tag name
   * @param label	label of the content
   */
  public void tag (String tag,
		   String label)
  {
    out.print ("<");
    out.print (tag);
    out.print (">");
    translate (label);
    out.print ("</");
    out.print (tag);
    out.println (">");
  }

  /** Generate markup for a link
   * @param name	label of the link name (null=use directory name)
   * @param target	label of the link target
   * @param directory	link target directory name (untranslated, null=none)
   * @param subtarget	link subtarget name (untranslated, null=none)
   */
  public void link (String name,
		    String target,
		    String directory,
		    String subtarget)
  {
    out.print ("<A HREF=\"");
    if (directory != null) {
      url (directory);
      out.print ('/');
    }
    url (bundle.getString (target));
    if (subtarget != null) {
      out.print ('#');
      url (subtarget);
    }
    out.print ("\">");
    if (name != null)
      translate (name);
    else
      html (directory);
    out.print ("</A>");
  }

  /** Generate markup for a link
   * @param name	the link name
   * @param target	label of the link target
   * @param directory	link target directory name (untranslated, null=none)
   * @param subtarget	link subtarget name (untranslated, null=none)
   */
  public void link (int name,
		    String target,
		    String directory,
		    String subtarget)
  {
    out.print ("<A HREF=\"");
    if (directory != null) {
      url (directory);
      out.print ('/');
    }
    url (bundle.getString (target));
    if (subtarget != null) {
      out.print ('#');
      url (subtarget);
    }
    out.print ("\">");
    out.print (name);
    out.print ("</A>");
  }

  /** Generate markup for an inlined image
   * @param directory	the directory name
   * @param filename	name of the image file
   */
  public void img (String directory, String filename)
  {
    out.print ("<A HREF=\"");
    if (!directory.equals ("")) {
      url (directory);
      out.print ('/');
    }
    url (filename);
    out.print ("\"><IMG ALIGN=MIDDLE ALT=\"");
    translate ("picture");
    out.print ("\" SRC=\"");
    if (!directory.equals ("")) {
      url (directory);
      out.print ('/');
    }
    url ("t/");
    url (filename);
    out.print ("\"></A>");
  }

  /** Generate markup for inlined images
   * @param directory	the directory name
   * @param basename	base name of the image file
   * @param suffixes	suffixes of the image files, separated by spaces
   */
  public void img (String directory, String basename, String suffixes)
  {
    int i = 0, j = suffixes.indexOf (' ');
    while (j >= 0) {
      img (directory, basename + suffixes.substring (i, j));
      i = j + 1;
      j = suffixes.indexOf (' ', i);
    }
    if (suffixes.length () > i)
      img (directory, basename + suffixes.substring (i));
  }

  /** Output a URL string
   * @param name	URL to be output
   */
  public void url (String name)
  {
    out.print (urlencode (name));
  }

  /** Output an HTML string
   * @param name	text to be output
   */
  public void html (String text)
  {
    out.print (htmlencode (text));
  }

  /** Encode a name as a URI
   * @param name	name to be encoded
   * @return		the name converted to a URI
   */
  public static String urlencode (String name)
  {
    StringBuffer encoded = new StringBuffer (name.length ());
    for (int i = 0; i < name.length (); i++) {
      final char b = name.charAt (i);
      if (b >= 0 && b < 16)
	encoded.append ("%0").append (Integer.toHexString (b));
      else if (b < ' ' || b >= 127 || " \"#$%&'+:<>?".indexOf (b) != -1)
	encoded.append ("%").append (Integer.toHexString (b));
      else
	encoded.append (b);
    }
    return encoded.toString ();
  }

  /** Encode a name as a HTML string
   * @param name	name to be encoded
   * @return		the name converted to HTML text
   */
  public static String htmlencode (String name)
  {
    StringBuffer encoded = new StringBuffer (name.length ());
    for (int i = 0; i < name.length (); i++) {
      switch (name.charAt (i)) {
      case '&':
	encoded.append ("&amp;");
	break;
      case '<':
	encoded.append ("&lt;");
	break;
      case '>':
	encoded.append ("&gt;");
	break;
      default:
	encoded.append (name.charAt (i));
      }
    }
    return encoded.toString ();
  }
}
