package results;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Vector;

/** Series of contests */
public final class Series
{
  /** The file name extension for series files */
  public final static String FILE_EXTENSION = ".pts";
  /** The file the series was last synchronized with */
  public File file;
  /** The constraints used by the series */
  public Constraints constraints = new Constraints ();
  /** Flag: is this a national series */
  boolean national;
  /** Name of the tournament */
  String tournament;
  /** Name of the series */
  String name;
  /** Teams in the series */
  Vector teams = new Vector ();
  /** Competitors in the series */
  Vector competitors = new Vector ();
  /** Judges in the series */
  Vector judges = new Vector ();
  /** Flag: has the series been modified */
  private transient boolean modified = false;
  /** Contests in the series */
  private Vector contests = new Vector ();
  /** The countries */
  public Countries countries;
  /** The locale for this series */
  Locale locale;
  
  /** Constructor
   * @param locale	Locale of the series
   */
  public Series (Locale locale)
  {
    countries = new Countries (this.locale = locale);
    tournament = Constraints.CONTEST_TITLE;
  }

  /** Convert the series to a printable string */
  public String toString () { return tournament + " - " + name; }

  /** Flag the series modified */
  public void flagModified () { modified = true; }
  /** Determine whether the series data has been modified */
  public boolean isModified () { return modified; }

  /** Flag all contests in the series modified */
  public void flagContestsModified () {
    for (Enumeration en = contests.elements (); en.hasMoreElements (); )
      ((Contest) en.nextElement ()).flagModified ();
  }

  /** Change the locale for this series
   * @param locale	the desired locale
   */
  public void setLocale (Locale locale)
  {
    if (this.locale != locale) {
      modified = true;
      countries.setLocale (this.locale = locale);
    }
  }

  /** Get the name of the series */
  public String getName () { return name; }
  /** Set the name of the series */
  public void setName (String name) {
    this.name = name;
    modified = true;
  }
  /** Get the name of the tournament */
  public String getTournament () { return tournament; }
  /** Get the number of judges */
  public int getNumJudges () { return judges.size (); }
  /** Get the nth judge */
  public Judge getJudge (int i) { return (Judge) judges.elementAt (i); }
  /** Get a collection of the judges */
  public Collection getJudges () { return judges; }
  /** Get the number of competitors */
  public int getNumCompetitors () {
    return competitors.size ();
  }
  /** Get the nth competitor */
  public Competitor getCompetitor (int i) {
    return (Competitor) competitors.elementAt (i);
  }
  /** Get a collection of the competitors */
  public Collection getCompetitors () { return competitors; }
  /** Get the number of teams */
  public int getNumTeams () {
    return teams.size ();
  }
  /** Get the nth team */
  public Team getTeam (int i) {
    return (Team) teams.elementAt (i);
  }
  /** Get a collection of all teams */
  public Collection getTeams () { return teams; }
  /** Get the number of contests */
  public int getNumContests () {
    return contests.size ();
  }
  /** Get the nth contest */
  public Contest getContest (int i) {
    return (Contest) contests.elementAt (i);
  }
  /** Get a collection of all contests */
  public Collection getContests () { return contests; }

  /** See if this is a national series */
  public boolean isNational () { return national; }
  /** Toggle the "national" status of the series
   * @return	the new status
   */
  public boolean toggleNational () {
    flagContestsModified ();
    return national = !national;
  }

  /** Get the locale for this series
   * @return	the locale
   */
  public Locale getLocale () { return locale; }

  /** Update the name of the series according to a file
   * @param file	the file
   */
  private void setName (File file)
  {
    this.file = file;
    name = file.getName ();
    int i = name.lastIndexOf (FILE_EXTENSION);
    if (i >= 0)
      name = name.substring (0, i);
  }

  /** Read the series from permanent storage
   * @param file	the input file
   * @exception		FileNotFoundException if the file cannot be found
   * @exception		IOException if the file cannot be read
   */
  public void read (File file)
    throws FileNotFoundException, IOException
  {
    read (new FileReader (file));
    setName (file);
  }

  /** Write the series to permanent storage
   * @param file	the output file
   * @exception		IOException if the file cannot be read
   */
  public void write (File file)
    throws IOException
  {
    write (new FileWriter (file));
    setName (file);
  }

  /** Read the series from permanent storage
   * @param reader	the input stream
   * @exception		IOException if the file cannot be read
   */
  private void read (Reader reader)
    throws IOException
  {
    LineNumberReader in = new LineNumberReader (reader);
    String line = in.readLine ();
    try {
      int i = line.indexOf (' ');
      if (i >= 0 && line.substring (0, i).equals ("national"))
	national = true;
      else if (i >= 0 && line.substring (0, i).equals ("international"))
	national = false;
      else
	throw new Exception ();
      setLocale (new Locale (line.substring (i + 1), ""));
      line = in.readLine ();
      if (line == null) throw new Exception ();
      tournament = line;
      line = in.readLine ();
      if (line.startsWith ("min ")) {
	i = line.indexOf (' ');
	for (int k = Constraints.MAX_POINTS + 1; k-- > 0; ) {
	  int j = line.indexOf (' ', i + 1);
	  constraints.min[k] = Integer.parseInt (line.substring (i + 1, j));
	  i = j;
	}
	line = in.readLine ();
      }
      if (line.startsWith ("max ")) {
	i = line.indexOf (' ');
	for (int k = Constraints.MAX_POINTS + 1; k-- > 0; ) {
	  int j = line.indexOf (' ', i + 1);
	  constraints.max[k] = Integer.parseInt (line.substring (i + 1, j));
	  i = j;
	}
	line = in.readLine ();
      }
      if (line.startsWith ("rect ")) {
	i = line.indexOf (' ');
	int j = line.indexOf (' ', i + 1);
	int nRect = Integer.parseInt (line.substring (i + 1, j));
	if (nRect < 0)
	  throw new Exception ("negative number of rectification rounds");
	else if (nRect == 0)
	  constraints.rect = null;
	else {
	  constraints.rect = new int[nRect];
	  for (int k = constraints.rect.length; k-- > 0; ) {
	    i = j;
	    j = line.indexOf (' ', i + 1);
	    int r = Integer.parseInt (line.substring (i + 1, j));
	    if (r < 0) throw new Exception ("negative rectification limit");
	    constraints.rect[k] = r;
	  }
	}
	line = in.readLine ();
      }
      if (line.equals ("judges")) {
	judges = new Vector ();
	for (;;) {
	  line = in.readLine ();
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  Country c = countries.get (line.substring (0, 2));
	  if (c == null) throw new Exception ("no such country");
	  add (new Judge (line.substring (i + 1), c));
	}
      }
      else
	throw new Exception ();
      if (line.equals ("competitors")) {
	competitors = new Vector ();
	for (;;) {
	  line = in.readLine ();
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  Country c = countries.get (line.substring (0, 2));
	  if (c == null) throw new Exception ("no such country");
	  add (new Competitor (line.substring (i + 1), c));
	}
      }
      else
	throw new Exception ();
      if (line.equals ("teams")) {
	teams = new Vector ();
	for (;;) {
	  line = in.readLine ();
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  int size = Integer.parseInt (line.substring (0, i));
	  Team team = new Team (this, null);
	  team.competitors = new Vector ();
	  while (size-- > 0) {
	    int j = line.indexOf (' ', i + 1);
	    int num = Integer.parseInt (line.substring (i + 1, j));
	    i = j;
	    if (num < 0 || num >= competitors.size ())
	      throw new Exception ("no competitor " + num);
	    Competitor comp = (Competitor) competitors.elementAt (num);
	    if (team.competitors.contains (comp))
	      throw new Exception ("duplicate competitor " + num);
	    if (comp.team != null)
	      throw new Exception ("duplicate team for competitor " + num);
	    comp.team = team;
	    team.competitors.addElement (comp);
	  }
	  team.name = line.substring (i + 1);
	  add (team);
	}
      }
      else
	throw new Exception ();
      if (line.equals ("contests")) {
	contests = new Vector ();
	for (;;) {
	  line = in.readLine ();
	  if (line == null) break;
	  i = line.indexOf (' ');
	  if (i < 0) break;
	  int numComp = Integer.parseInt (line.substring (0, i));
	  int j = line.indexOf (' ', i + 1);
	  int numJudges = Integer.parseInt (line.substring (i + 1, j));
	  Vector jury = new Vector (numJudges);
	  i = j;
	  for (; numJudges-- > 0; i = j) {
	    j = line.indexOf (' ', i + 1);
	    int num = Integer.parseInt (line.substring (i + 1, j));
	    if (num < 0 || num > judges.size ())
	      throw new Exception ("no judge " + num);
	    jury.addElement (judges.elementAt (num));
	  }
	  final String format = line.substring (i + 1);
	  Contest contest =
	    new Contest (line = in.readLine (), jury);
	  contest.setFormat (format);
	  add (contest);
	  while (numComp-- > 0) {
	    line = in.readLine ();
	    i = line.indexOf (' ');
	    int num = Integer.parseInt (line.substring (0, i));
	    if (num < 0 || num > competitors.size ())
	      throw new Exception ("no competitor " + num);
	    Competitor comp = (Competitor) competitors.elementAt (num);
	    j = line.indexOf (' ', i + 1);
	    num = Integer.parseInt (line.substring (i + 1, j));
	    Mirror mirror = new Mirror (comp, contest, num);
	    i = j;
	    j = line.indexOf (' ', i + 1);
	    mirror.penalty = Integer.parseInt (line.substring (i + 1, j));
	    if (mirror.penalty != 0) {
	      i = j;
	      j = line.indexOf (' ', i + 1);
	      mirror.penaltyReason = line.charAt (j - 1);
	    }
	    Integer points[] = new Integer[jury.size ()];
	    for (numJudges = 0; numJudges < jury.size (); numJudges++) {
	      i = j;
	      j = line.indexOf (' ', i + 1);
	      if (line.charAt (j - 1) != '*')
		points[numJudges] =
		  new Integer (Integer.parseInt (line.substring (i + 1, j)));
	    }
	    mirror.setPoints (points);
	    contest.add (mirror);
	  }
	}
	if ((line = in.readLine ()) != null)
	  throw new Exception ();
      }
      else
	throw new Exception ();
    }
    catch (IOException e) {
      throw e;
    }
    catch (Exception e) {
      String msg = e.toString ();
      if (msg == null)
	msg = "unrecognized input";
      if (line == null)
	line = "(end of file)";
      else
	line = "'" + line + "'";
      throw new IOException (name + ":" + in.getLineNumber () +
			     ": " + msg + ": " + line);
    }
    modified = false;
  }

  /** Write the series to permanent storage
   * @param writer	the output stream
   * @exception		IOException if the file cannot be written
   */
  private void write (Writer writer)
    throws IOException
  {
    PrintWriter out = new PrintWriter (writer);
    Enumeration en;
    int i;
    out.print (national ? "national " : "international ");
    out.println (locale.getLanguage ());
    out.println (tournament);
    out.print ("min");
    for (i = Constraints.MAX_POINTS + 1; i-- > 0; )
      out.print (" " + constraints.min[i]);
    out.println (" ");
    out.print ("max");
    for (i = Constraints.MAX_POINTS + 1; i-- > 0; )
      out.print (" " + constraints.max[i]);
    out.println (" ");
    out.print ("rect");
    if (constraints.rect != null) {
      out.print (" " + constraints.rect.length);
      for (i = constraints.rect.length; i-- > 0; )
	out.print (" " + constraints.rect[i]);
    }
    else
      out.print (" 0");
    out.println (" ");
    out.println ("judges");
    for (en = judges.elements (); en.hasMoreElements (); ) {
      Judge judge = (Judge) en.nextElement ();
      out.println (judge.country.code + " " + judge.name);
    }
    out.println ("competitors");
    for (en = competitors.elements (); en.hasMoreElements (); ) {
      Competitor competitor = (Competitor) en.nextElement ();
      out.println (competitor.country.code + " " + competitor.name);
    }
    out.println ("teams");
    for (en = teams.elements (); en.hasMoreElements (); ) {
      Team team = (Team) en.nextElement ();
      out.print (team.size ());
      for (Enumeration enm = team.elements (); enm.hasMoreElements (); )
	out.print (" " + competitors.indexOf (enm.nextElement ()));
      out.print (" ");
      out.println (team.name);
    }
    out.println ("contests");
    for (en = contests.elements (); en.hasMoreElements (); ) {
      Contest contest = (Contest) en.nextElement ();
      out.print (contest.mirrors.size ());
      out.print (" ");
      out.print (contest.judges.size ());
      Enumeration enm;
      for (enm = contest.judges.elements (); enm.hasMoreElements (); )
	out.print (" " + judges.indexOf (enm.nextElement ()));
      out.print (" ");
      out.println (contest.getFormat () == null ? "" : contest.getFormat ());
      out.println (contest.name);
      for (enm = contest.mirrors.elements (); enm.hasMoreElements (); ) {
	Mirror mirror = (Mirror) enm.nextElement ();
	out.print (competitors.indexOf (mirror.competitor) + " " +
		   mirror.number + " " + mirror.penalty);
	if (mirror.penalty != 0)
	  out.print (" " + mirror.penaltyReason);
	Integer points[] = mirror.getPoints ();
	for (i = 0; i < points.length; i++)
	  out.print (points[i] == null ? " *" : " " + points[i]);
	out.println (" ");
      }
    }
    out.close ();
    modified = false;
  }

  /** Add a judge to the series
   * @param judge	the judge to be added
   */
  public void add (Judge judge) {
    judge.series = this;
    judges.add (judge);
    modified = true;
  }

  /** Add a competitor to the series
   * @param competitor	the competitor to be added
   */
  public void add (Competitor competitor) {
    competitor.series = this;
    competitors.add (competitor);
    modified = true;
  }

  /** Add a team to the series
   * @param team	the team to be added
   */
  public void add (Team team) {
    team.series = this;
    teams.add (team);
    modified = true;
  }

  /** Add a contest to the series
   * @param contest	the contest to be added
   */
  public void add (Contest contest) {
    contest.series = this;
    contests.add (contest);
    modified = true;
  }

  /** Remove a judge from the series
   * @param judge	the judge to be removed
   * @return		true if the judge belonged to the series
   */
  public boolean remove (Judge judge)
  {
    if (!judges.remove (judge))
      return false;
    for (Enumeration en = contests.elements (); en.hasMoreElements (); )
      ((Contest) en.nextElement ()).remove (judge);
    return modified = true;
  }

  /** Remove a competitor from the series
   * @param competitor	the competitor to be removed
   * @return		true if the competitor belonged to the series
   */
  public boolean remove (Competitor competitor)
  {
    if (!competitors.remove (competitor))
      return false;
    if (competitor.team != null)
      competitor.team.remove (competitor);
    for (Enumeration en = contests.elements (); en.hasMoreElements (); )
      ((Contest) en.nextElement ()).remove (competitor);
    return modified = true;
  }

  /** Remove a team from the series
   * @param team	the team to be removed
   * @return		true if the team belonged to the series
   */
  public boolean remove (Team team)
  {
    if (!teams.remove (team))
      return false;
    team.clear ();
    return modified = true;
  }

  /** Remove a contest from the series
   * @param contest	the contest to be removed
   * @return		true if the contest belonged to the series
   */
  public boolean remove (Contest contest)
  {
    if (!contests.remove (contest))
      return false;
    return modified = true;
  }

  /** Generate the HTML output
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generate (ErrorReporter reporter)
    throws IOException
  {
    File dir = mkdir (file.getParentFile (), name);
    HTMLWriter html = new HTMLWriter (dir, "index.html", locale);
    html.header (toString (), "series.index");

    html.out.print ("<H2>");
    html.link ("mirrors.index", "mirrors.html", null, null);
    html.out.println ("</H2>");

    html.out.print ("<H2>");
    html.link ("results.index", "results.html", null, null);
    html.out.println ("</H2>");

    html.out.print ("<H2>");
    html.link ("teams.index", "teams.html", null, null);
    html.out.println ("</H2>");

    html.out.print ("<H2>");
    html.link ("awards.index", "awards.html", null, null);
    html.out.println ("</H2>");

    html.footer ();

    generateMirrors (dir, reporter);
    generateResults (dir, reporter);
    generateAwards (dir, 5, reporter);
    generateTeams (dir);
  }

  /** Generate an index of mirror numbers in the series
   * @param dir		the directory where the index is to be created
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateMirrors (File dir, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "mirrors.html", locale);
    html.header (toString (), "mirrors.index");
    html.out.println ("<OL>");

    for (Enumeration e = contests.elements (); e.hasMoreElements (); ) {
      Contest contest = (Contest) e.nextElement ();
      html.out.print ("<LI>");
      html.link (null, "mirrors.html", contest.name, null);
      html.out.println ("</LI>");
      contest.generateMirrors (mkdir (dir, contest.name), reporter);
    }

    html.out.println ("</OL>");
    html.footer ();
  }

  /** Generate individual results for the series
   * @param dir		the directory where the file is to be created
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateResults (File dir, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "results.html", locale);
    html.header (toString (), "results.index");

    html.out.println ("<TABLE BORDER>\n<TR>");
    html.tag ("TD", "results.rank");
    html.tag ("TD", "results.name");
    html.tag ("TD", "results.team");
    html.tag ("TD", "results.total");

    for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
      Contest contest = (Contest) c.nextElement ();
      html.out.print ("<TD>");
      html.link (null, "results.html", contest.name, null);
      html.out.println ("</TD>");
      contest.generateResults (mkdir (dir, contest.name), reporter);
    }
    html.out.println ("</TR>\n");

    // Display the rankings
    int place = 0, rank = 0, lasttotal = -1;
    for (Enumeration e = getRanking (); e.hasMoreElements (); place++) {
      Competitor competitor = (Competitor) e.nextElement ();

      int total = getPoints (competitor);
      if (total != lasttotal) {
	rank = place + 1;
	lasttotal = total;
      }

      html.out.println ("<TR>\n <TD ALIGN=RIGHT>" + rank + "</TD>\n <TD>" +
			competitor.name + "</TD>\n <TD>" +
			(competitor.team == null
			 ? competitor.country.name
			 : competitor.team.name + ", " +
			 competitor.country.name) + "</TD>\n <TD>" +
			total + "</TD>");

      for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
	Contest contest = (Contest) c.nextElement ();
	Mirror mirror = contest.getMirror (competitor);
	if (mirror != null) {
	  html.out.print ("<TD>" + mirror.rank + ": ");
	  html.link (mirror.getTotalRectified (), "results.html",
		     contest.name, Integer.toString (mirror.number));
	  if (contest.getFormat () != null) {
	    html.out.print (' ');
	    html.img (contest.name, new Integer (mirror.number).toString (),
		      contest.getFormat ());
	  }
	  html.out.println ("</TD>");
	}
	else
	  html.out.println ("<TD>-</TD>");
      }

      html.out.println ("</TR>");
    }

    html.out.println ("</TABLE>");
    html.footer ();
  }

  /** Generate award list for the series
   * @param dir		the directory where the file is to be created
   * @param top		number of top ranks to list
   * @param reporter	a reporter object for diagnostics
   * @exception		IOException if the output cannot be created
   */
  public void generateAwards (File dir, int top, ErrorReporter reporter)
    throws IOException
  {
    HTMLWriter html = new HTMLWriter (dir, "awards.html", locale);
    html.header (toString (), "awards.index");

    html.out.println ("<H2>" + name + "</H2>\n");

    for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
      Contest contest = (Contest) c.nextElement ();
      html.out.print ("<H3>" + name + "/" + contest.name + "</H3>\n");
      contest.generateAwards (html, top, reporter);
    }

    html.tag ("H3", "awards.total");

    // Display the rankings
    int place = 0, rank = 0, lasttotal = -1;
    for (Enumeration e = getRanking (); e.hasMoreElements (); place++) {
      Competitor competitor = (Competitor) e.nextElement ();

      int total = getPoints (competitor);
      if (total != lasttotal) {
	rank = place + 1;
	lasttotal = total;
      }

      if (rank > top)
	break;

      html.out.print ("<P>" + rank + "." +
		      competitor.name + "; " +
		      (competitor.team == null
		       ? competitor.country.name
		       : competitor.team.name + ", " +
		       competitor.country.name) + "(");

      for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
	Contest contest = (Contest) c.nextElement ();
	Mirror mirror = contest.getMirror (competitor);
	if (mirror != null)
	  html.out.print (mirror.getTotalRectified ());
	else
	  html.out.print ("?");
	if (c.hasMoreElements ())
	  html.out.print ("+");
      }

      html.out.println ("=" + total + ")</P>");
    }

    html.footer ();
  }

  /** Generate team results for the series
   * @param dir		the directory where the file is to be created
   * @exception		IOException if the output cannot be created
   */
  public void generateTeams (File dir)
    throws IOException
  {
    /** Minimum number of participants in a team */
    int numCompetitors = Integer.MAX_VALUE;
    for (int i = teams.size (); i-- > 0; ) {
      int size = ((Team) teams.elementAt (i)).size ();
      if (size < numCompetitors)
	numCompetitors = size;
    }

    HTMLWriter html = new HTMLWriter (dir, "teams.html", locale);
    html.header (toString (), "teams.index");

    html.out.println ("<TABLE BORDER>\n<TR>");
    html.tag ("TD", "results.rank");
    html.tag ("TD", "results.team");
    html.tag ("TD", "results.total");
    for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
      Contest contest = (Contest) c.nextElement ();
      html.out.print ("<TD>");
      html.link (null, "results.html", contest.name, null);
      html.out.println ("</TD>");
    }
    html.out.println ("</TR>");

    // Display the rankings
    int place = 0, rank = 0, lasttotal = -1;
    for (Enumeration e = getTeamRanking (numCompetitors);
	 e.hasMoreElements (); place++) {
      Object[] data = (Object[]) e.nextElement ();
      Team team = (Team) data[0];
      int points[] = (int[]) data[1];

      if (points[0] != lasttotal) {
	rank = place + 1;
	lasttotal = points[0];
      }

      html.out.print ("<TR>\n<TD ALIGN=RIGHT>" + rank + "</TD>\n<TD>");
      html.html (team.name);
      html.out.println ("</TD>\n<TD ALIGN=RIGHT>" + points[0] + "</TD>");

      {
	int i = 0;
	for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
	  Contest contest = (Contest) c.nextElement ();
	  html.out.println ("<TD ALIGN=RIGHT>" + points[++i] + "</TD>");
	}
      }

      html.out.println ("</TR>");
      for (Enumeration comp = team.elements (); comp.hasMoreElements (); ) {
	Competitor competitor = (Competitor) comp.nextElement ();
	html.out.print ("<TR>\n<TD></TD>\n<TD>");
	html.html ("" + competitor.name);
	html.out.println ("<TD ALIGN=RIGHT>" + getPoints (competitor) +
			  "</TD>");
	for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
	  Contest contest = (Contest) c.nextElement ();
	  Mirror mirror = contest.getMirror (competitor);
	  if (mirror != null) {
	    html.out.print ("<TD>" + mirror.rank + ": ");
	    html.link (mirror.getTotalRectified (), "results.html",
		       contest.name, Integer.toString (mirror.number));
	    if (contest.getFormat () != null) {
	      html.out.print (' ');
	      html.img (contest.name, new Integer (mirror.number).toString (),
			contest.getFormat ());
	    }
	    html.out.println ("</TD>");
	  }
	  else
	    html.out.println ("<TD></TD>");
	}
      }
      html.out.println ("</TR>");
    }

    html.out.println ("</TABLE>");
    html.footer ();
  }

  /** Determine the rankings for the series
   * @return	Enumeration to a vector of competitors in the ranking order
   */
  public Enumeration getRanking ()
  {
    Vector ranking = new Vector ();
    for (Enumeration e = competitors.elements (); e.hasMoreElements (); ) {
      Competitor competitor = (Competitor) e.nextElement ();
      int total = getPoints (competitor);
      int i;
      for (i = 0; i < ranking.size (); i++) {
	if (total > getPoints ((Competitor) ranking.elementAt (i)))
	  break;
      }
      if (i == ranking.size ())
	ranking.addElement (competitor);
      else
	ranking.insertElementAt (competitor, i);
    }
    return ranking.elements ();
  }

  /** Determine the rankings for the teams
   * @param numCompetitors	number of team members to consider
   * @return	Enumeration to an Object[] (Team,int[])
   *		where the first int are the total points
   *		and subsequent ints are the points for each contest
   */
  public Enumeration getTeamRanking (int numCompetitors)
  {
    Vector tr = new Vector ();
    for (Enumeration e = teams.elements (); e.hasMoreElements (); ) {
      Team team = (Team) e.nextElement ();
      int points[] = getPoints (team, numCompetitors);
      int i;
      for (i = 0; i < tr.size (); i++)
	if (points[0] > ((int[]) ((((Object[]) tr.elementAt (i)))[1]))[0])
	  break;
      Object obj[] = new Object[] {team, points};
      if (i == tr.size ())
	tr.addElement (obj);
      else
	tr.insertElementAt (obj, i);
    }
    return tr.elements ();
  }

  /** Determine the total points of a contestant in this series
   * @param competitor		the contestant
   * @return			total number of rectified points minus penalty
   */
  public int getPoints (Competitor competitor)
  {
    int total = 0;
    for (Enumeration c = contests.elements (); c.hasMoreElements (); ) {
      Contest contest = (Contest) c.nextElement ();
      Mirror mirror = contest.getMirror (competitor);
      if (mirror != null)
	total += mirror.getTotalRectified ();
    }
    return total;
  }

  /** Determine the total points of a team in this series
   * @param team		the team
   * @param numCompetitors	number of team members to consider
   * @return			total number of rectified points minus penalty,
   *				plus total points in each contest
   */
  public int[] getPoints (Team team, int numCompetitors)
  {
    int total[] = new int[1 + contests.size ()];
    for (int c = contests.size (); c-- > 0; ) {
      Contest contest = (Contest) contests.elementAt (c);
      int points[] = new int[numCompetitors];
      for (Enumeration comp = team.elements (); comp.hasMoreElements (); ) {
	Competitor competitor = (Competitor) comp.nextElement ();
	final Mirror mirror = contest.getMirror (competitor);
	final int p = mirror == null ? 0 : mirror.getTotalRectified ();
	for (int i = 0; i < numCompetitors; i++) {
	  if (points[i] < p) {
	    for (int j = numCompetitors; --j > i; )
	      points[j] = points[j - 1];
	    points[i] = p;
	    break;
	  }
	}
      }
      for (int i = numCompetitors; i-- > 0; )
	total[1 + c] += points[i];
      total[0] += total[1 + c];
    }
    return total;
  }

  /** Create a directory
   * @param parent	the parent directory
   * @param directory	the directory name
   * @exception		IOException if the directory cannot be created
   * @return		a file object to the directory
   */
  public static File mkdir (File parent, String directory)
    throws IOException
  {
    File file = new File (parent, directory);
    if (!file.isDirectory () && !file.mkdir ())
      throw new IOException ("mkdir " + file);
    return file;
  }
}
