package results.ui;

import results.Country;
import results.Competitor;
import results.Team;

import java.text.Collator;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Locale;
import java.util.TreeSet;
import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JTable;
import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

/** Table for viewing and editing competitor data */
final class CompetitorTable
  extends JTable
{
  /** The series dialog containing this table */
  private SeriesDialog parent;
  /** The country box */
  private CountryBox countrybox;
  /** The country cell renderer */
  private DefaultTableCellRenderer rcountry;
  /** The competitor name cell renderer */
  private DefaultTableCellRenderer rcompetitor;
  /** The competitor team cell renderer */
  private DefaultTableCellRenderer rteam;

  /** The team box */
  private TeamBox teambox;
  /** The empty team */
  private final static Team empty = new Team (null, "");

  /** The country column number */
  private static final int COLUMN_COUNTRY = 0;
  /** The competitor name column number */
  private static final int COLUMN_COMPETITOR = 1;
  /** The team name column number */
  private static final int COLUMN_TEAM = 2;

  /** Constructor
   * @param dialog	the containing series dialog
   */
  CompetitorTable (SeriesDialog dialog)
  {
    parent = dialog;
    setModel (new AbstractTableModel ()
      {
	public int getColumnCount () { return 3; }
	public int getRowCount () {
	  return parent == null ? 1 : parent.series.getNumCompetitors () + 1;
	}
	public boolean isCellEditable (int row, int col) {
	  return true;
	}
	public Class getColumnClass (int col) {
	  switch (col) {
	  case COLUMN_COUNTRY:
	    return Country.class;
	  case COLUMN_COMPETITOR:
	    return String.class;
	  case COLUMN_TEAM:
	    return Team.class;
	  }
	  return null;
	}
	public Object getValueAt (int row, int col) {
	  final int rows = getRowCount ();
	  if (row < rows - 1) {
	    final Competitor competitor = parent.series.getCompetitor (row);
	    switch (col) {
	    case COLUMN_COUNTRY:
	      return competitor.getCountry ();
	    case COLUMN_COMPETITOR:
	      return competitor.getName ();
	    case COLUMN_TEAM:
	      return competitor.getTeam ();
	    }
	  }
	  return null;
	}
	public void setValueAt (Object value, int row, int col) {
	  final int rows = getRowCount ();
	  Competitor competitor;
	  if (row < rows - 1)
	    competitor = parent.series.getCompetitor (row);
	  else {
	    competitor = new Competitor ("", countrybox.getDefault ());
	    parent.series.add (competitor);
	  }
	  switch (col) {
	  case COLUMN_COUNTRY:
	    if (value != null)
	      competitor.setCountry ((Country) value);
	    break;
	  case COLUMN_COMPETITOR:
	    if (value == null || value.equals ("")) {
	      if (row >= rows - 1 ||
		  parent.confirm ("competitor.remove", competitor.getName ()))
		parent.series.remove (competitor);
	    }
	    else
	      competitor.setName ((String) value);
	    break;
	  case COLUMN_TEAM:
	    Team team = competitor.getTeam ();
	    if (team != null)
	      team.remove (competitor);
	    if (value != null && value != empty)
	      ((Team) value).add (competitor);
	    competitor.setTeam (value == empty ? null : (Team) value);
	    break;
	  }
	}
	public String getColumnName (int col) {
	  switch (col) {
	  case COLUMN_COUNTRY:
	    return parent.translator.lookup ("competitor.nCountry");
	  case COLUMN_COMPETITOR:
	    return parent.translator.lookup ("competitor.nCompetitor");
	  case COLUMN_TEAM:
	    return parent.translator.lookup ("competitor.nTeam");
	  }
	  return null;
	}
      });
    countrybox = new CountryBox (parent.series.countries);
    setDefaultEditor (Country.class, new DefaultCellEditor (countrybox));
    setDefaultEditor
      (Team.class, new DefaultCellEditor (teambox = new TeamBox ()));
    setDefaultRenderer (Country.class,
			rcountry = new DefaultTableCellRenderer ());
    setDefaultRenderer (String.class,
			rcompetitor = new DefaultTableCellRenderer ());
    setDefaultRenderer (Team.class,
			rteam = new DefaultTableCellRenderer ());
  }

  public void setLocale (Locale locale)
  {
    super.setLocale (locale);
    countrybox.updateCountries ();
    rcountry.setToolTipText
      (parent.translator.lookup ("competitor.tCountry"));
    rcompetitor.setToolTipText
      (parent.translator.lookup ("competitor.tCompetitor"));
    rteam.setToolTipText
      (parent.translator.lookup ("competitor.tTeam"));
    tableChanged (new TableModelEvent
      (getModel (), TableModelEvent.HEADER_ROW));
  }

  /** Update the data on available teams */
  public void updateTeams () {
    teambox.updateTeams ();
  }

  /** A combo box for picking a team */
  final class TeamBox
    extends JComboBox
  {
    /** The default team */
    private Team team;
    TeamBox () {
      updateTeams ();
    }
    /** Rebuild the team list */
    void updateTeams () {
      TreeSet set = new TreeSet (new Comparator ()
	{
	  final Collator collator =
	    Collator.getInstance (CompetitorTable.this.parent.getLocale ());
	  public int compare (Object o1, Object o2) {
	    return collator.compare (((Team) o1).getName (),
				     ((Team) o2).getName ());
	  }
	});
      set.addAll (CompetitorTable.this.parent.series.getTeams ());
      removeAllItems ();
      addItem (CompetitorTable.empty);
      for (Iterator i = set.iterator (); i.hasNext (); )
	addItem (i.next ());
    }
  }
}
