package results.ui;

import results.Competitor;
import results.Constraints;
import results.Contest;
import results.Judge;
import results.Mirror;
import results.Reason;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.text.Collator;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Locale;
import java.util.TreeSet;
import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

/** Table for viewing and editing mirror data */
final class MirrorTable
  extends JTable
{
  /** The series dialog containing this table */
  private SeriesDialog parent;
  /** The controlling contest table */
  private ContestTable contest;
  /** The competitor cell renderer */
  private DefaultTableCellRenderer rcompetitor;
  /** The mirror cell renderer */
  private DefaultTableCellRenderer rmirror;
  /** The penalty cell renderer */
  private DefaultTableCellRenderer rpenalty;
  /** The penalty reason cell renderer */
  private DefaultTableCellRenderer rreason;
  /** The judge cell renderer */
  private DefaultTableCellRenderer rjudge;

  /** The popup menu for editing judges */
  private JPopupMenu judgepopup;
  /** The menu item "prioritize judge" */
  private JMenuItem judge_higher;
  /** The menu item "deprioritize judge" */
  private JMenuItem judge_lower;
  /** The submenu "add judge" */
  private JMenu judge_add;
  /** The menu entry "remove judge" */
  private JMenuItem judge_remove;
  /** The judge popup menu listener */
  private JudgeListener jl = new JudgeListener ();

  /** The popup menu for editing competitors */
  private JPopupMenu competitorpopup;
  /** The menu item "remove competitors" */
  private JMenuItem competitor_clear;
  /** The menu item "add all competitors" */
  private JMenuItem competitor_add;

  /** The popup menu for editing mirror numbers */
  private JPopupMenu mirrorpopup;
  /** The menu item "clear numbers" */
  private JMenuItem mirror_clear;
  /** The menu item "randomize numbers" */
  private JMenuItem mirror_random;

  /** Mouse listener for the table header */
  private MouseAdapter headerlistener;
  /** The selected header column for headerlistener */
  private int headercolumn;

  /** The reason box */
  private ReasonBox reasonbox;
  /** The competitor box */
  private CompetitorBox competitorbox;
  /** The empty competitor */
  private final static Competitor empty = new Competitor ("", null);

  /** A hash table of too frequent points given by different judges */
  private Hashtable frequent = new Hashtable ();

  /** The competitor column number */
  private static final int COLUMN_COMPETITOR = 0;
  /** The mirror number column number */
  private static final int COLUMN_MIRROR = 1;
  /** The penalty points column number */
  private static final int COLUMN_PENALTY = 2;
  /** The penalty reason column number */
  private static final int COLUMN_REASON = 3;
  /** The first judge column number */
  private static final int COLUMN_JUDGE = 4;

  /** Constructor
   * @param dialog	the containing series dialog
   * @param contest	the controlling contest table
   */
  MirrorTable (SeriesDialog dialog, ContestTable contest)
  {
    parent = dialog;
    this.contest = contest;
    setModel (new AbstractTableModel ()
      {
	public int getColumnCount () {
	  final Contest c = MirrorTable.this.contest.getSelected ();
	  final int numJudges = c == null ? 0 : c.getNumJudges ();
	  return COLUMN_JUDGE + (numJudges == 0 ? 1 : numJudges);
	}
	public int getRowCount () {
	  final Contest c = MirrorTable.this.contest.getSelected ();
	  return c == null ? 1 : c.getNumMirrors () + 1;
	}
	public boolean isCellEditable (int row, int col) {
	  final Contest c = MirrorTable.this.contest.getSelected ();
	  final int numJudges = c == null ? 0 : c.getNumJudges ();
	  return c != null && (col == 0 || (row < getRowCount () - 1 &&
					    col < COLUMN_JUDGE + numJudges));
	}
	public Class getColumnClass (int col) {
	  switch (col) {
	  case COLUMN_COMPETITOR:
	    return Competitor.class;
	  case COLUMN_REASON:
	    return Reason.class;
	  default:
	    return Integer.class;
	  }
	}
	public Object getValueAt (int row, int col) {
	  final int rows = getRowCount ();
	  if (row < rows - 1) {
	    final Contest c = MirrorTable.this.contest.getSelected ();
	    final Mirror mirror = c.getMirror (row);
	    switch (col) {
	    case COLUMN_COMPETITOR:
	      return mirror.getCompetitor ();
	    case COLUMN_MIRROR:
	      return mirror.getNumber ();
	    case COLUMN_PENALTY:
	      return mirror.getPenalty ();
	    case COLUMN_REASON:
	      return mirror.getPenalty () == null
		? null
		: reasonbox.getReason (mirror.getPenaltyReason ());
	    case COLUMN_JUDGE:
	    default:
	      return c == null || c.getNumJudges () == 0
		? null
		: mirror.getPoints (c.getJudge (col - COLUMN_JUDGE));
	    }
	  }
	  return null;
	}
	public void setValueAt (Object value, int row, int col) {
	  final Contest c = MirrorTable.this.contest.getSelected ();
	  if (c == null)
	    return;
	  final int rows = getRowCount ();
	  if (row < rows - 1) {
	    Mirror mirror = c.getMirror (row);
	    switch (col) {
	    case COLUMN_COMPETITOR:
	      if (value == empty) value = null;
	      if (value == null) {
		if (!parent.confirm ("mirror.competitor.remove",
				     mirror.getCompetitor ().toString ()))
		  break;
		c.remove (mirror);
		tableChanged
		  (new TableModelEvent (this, row, row,
					TableModelEvent.ALL_COLUMNS,
					TableModelEvent.DELETE));
	      }
	      else if (value == mirror.getCompetitor ())
		break;
	      else {
		if ((mirror.getNumber () != null ||
		     mirror.getPenalty () != null ||
		     mirror.hasPoints ()) &&
		    !parent.confirm ("mirror.competitor.change",
				     mirror.getCompetitor ().toString ()))
		  break;
		mirror.setCompetitor ((Competitor) value);
	      }
	      updateCompetitors ();
	      break;
	    case COLUMN_MIRROR:
	      if ((value == null ? 0 : ((Integer) value).intValue ()) ==
		  mirror.getNum ())
		break;
	      if ((mirror.getPenalty () != null || mirror.hasPoints ()) &&
		  !parent.confirm ("mirror.mirror",
				   mirror.getCompetitor ().toString ()))
		break;
	      if (value != null && ((Integer) value).intValue () < 0) {
		parent.warn ("mirror.mirror.positive",
			     mirror.getCompetitor ().toString ());
		break;
	      }
	      /** flag: was this previously a duplicate mirror number? */
	      final boolean dup = c.getNumOccurrences ((Integer) value) > 1;
	      mirror.setNumber ((Integer) value);
	      if (dup != c.getNumOccurrences ((Integer) value) > 1)
		tableChanged
		  (new TableModelEvent (this, 0, rows, COLUMN_MIRROR));
	      break;
	    case COLUMN_PENALTY:
	      if (value != null) {
		if (((Integer) value).intValue () < 0) {
		  parent.warn ("mirror.penalty.positive",
			       mirror.getCompetitor ().toString ());
		  break;
		}
		if (mirror.getPenaltyReason () == 0)
		  mirror.setPenaltyReason (reasonbox.getDefault ());
	      }
	      mirror.setPenalty ((Integer) value);
	      tableChanged
		(new TableModelEvent (this, row, row, COLUMN_REASON));
	      break;
	    case COLUMN_REASON:
	      if (value != null) {
		mirror.setPenaltyReason ((Reason) value);
		if (mirror.getPenalty () == null)
		  mirror.setPenalty (new Integer (1));
	      }
	      else
		mirror.setPenalty (null);
	      tableChanged
		(new TableModelEvent (this, row, row, COLUMN_PENALTY));
	      break;
	    default:
	      if (c.getNumJudges () == 0)
		break;
	      Judge judge = c.getJudge (col - COLUMN_JUDGE);
	      if (value == null)
		mirror.setPoints (judge, null);
	      else if (((Integer) value).intValue () < 0 ||
		       ((Integer) value).intValue () > Constraints.MAX_POINTS)
		parent.warn ("mirror.points",
			     mirror.getCompetitor ().toString (),
			     new Object[] {
			       new Integer (Constraints.MAX_POINTS)
			     });
	      else {
		mirror.setPoints (judge, (Integer) value);
		frequent.remove (judge);
		tableChanged (new TableModelEvent (this, 0, rows, col));
	      }
	      break;
	    }
	  }
	  else if (col == 0 && value != null && value != empty) {
	    c.add ((Competitor) value);
	    updateCompetitors ();
	  }
	}
	public String getColumnName (int col) {
	  switch (col) {
	  case COLUMN_COMPETITOR:
	    return parent.translator.lookup ("mirror.nCompetitor");
	  case COLUMN_MIRROR:
	    return parent.translator.lookup ("mirror.nMirror");
	  case COLUMN_PENALTY:
	    return parent.translator.lookup ("mirror.nPenalty");
	  case COLUMN_REASON:
	    return parent.translator.lookup ("mirror.nReason");
	  case COLUMN_JUDGE:
	  default:
	    final Contest c = MirrorTable.this.contest.getSelected ();
	    return c == null || c.getNumJudges () == 0
	      ? parent.translator.lookup ("mirror.nJudge")
	      : c.getJudge (col - COLUMN_JUDGE).getName ();
	  }
	}
      });

    ActionListener al;
    al = new CompetitorListener ();
    competitorpopup = new JPopupMenu ();
    competitorpopup.add (competitor_clear = new JMenuItem ());
    competitorpopup.add (competitor_add = new JMenuItem ());
    competitor_clear.addActionListener (al);
    competitor_add.addActionListener (al);

    al = new MirrorListener ();
    mirrorpopup = new JPopupMenu ();
    mirrorpopup.add (mirror_clear = new JMenuItem ()).addActionListener (al);
    mirrorpopup.add (mirror_random = new JMenuItem ()).addActionListener (al);

    judgepopup = new JPopupMenu ();
    judgepopup.add (judge_higher = new JMenuItem ()).addActionListener (jl);
    judgepopup.add (judge_lower = new JMenuItem ()).addActionListener (jl);
    judgepopup.add (new JPopupMenu.Separator ());
    judgepopup.add (judge_add = new JMenu ());
    judgepopup.add (judge_remove = new JMenuItem ()).addActionListener (jl);

    headerlistener = new MouseAdapter () {
        final public void mousePressed (MouseEvent e) {
	  headercolumn = getColumnModel ().getColumnIndexAtX (e.getX ());
	  final Contest c = MirrorTable.this.contest.getSelected ();
	  if (c == null);
	  else if (headercolumn == COLUMN_COMPETITOR)
	    competitorpopup.show (e.getComponent (), e.getX (), e.getY ());
	  else if (headercolumn == COLUMN_MIRROR && c.getNumMirrors () > 0)
	    mirrorpopup.show (e.getComponent (), e.getX (), e.getY ());
	  else if (headercolumn >= COLUMN_JUDGE) {
	    int numJudges = c.getNumJudges ();
	    judge_lower.setEnabled (headercolumn <
				    COLUMN_JUDGE + numJudges - 1);
	    judge_higher.setEnabled (headercolumn > COLUMN_JUDGE);
	    judge_remove.setEnabled (numJudges > 0);
	    judge_add.removeAll ();
	    for (Iterator it = parent.series.getJudges ().iterator ();
		 it.hasNext (); ) {
	      Judge judge = (Judge) it.next ();
	      if (!c.hasJudge (judge))
		judge_add.add (judge.toString ()).addActionListener (jl);
	    }
	    judgepopup.show (e.getComponent (), e.getX (), e.getY ());
	  }
        }
      };

    setDefaultEditor (Competitor.class, new DefaultCellEditor
      (competitorbox = new CompetitorBox ()));
    setDefaultRenderer (Competitor.class,
			rcompetitor = new DefaultTableCellRenderer ());
    setDefaultEditor (Reason.class, new DefaultCellEditor
      (reasonbox = new ReasonBox ()));
    setDefaultRenderer (Reason.class,
			rreason = new DefaultTableCellRenderer ());
    rmirror = new MirrorRenderer ();
    rpenalty = new DefaultTableCellRenderer ();
    setDefaultRenderer (Integer.class,
			rjudge = new ScoreRenderer ());
  }

  public void setLocale (Locale locale)
  {
    super.setLocale (locale);
    reasonbox.setLocale (locale);
    rcompetitor.setToolTipText
      (parent.translator.lookup ("mirror.tCompetitor"));
    rmirror.setToolTipText (parent.translator.lookup ("mirror.tMirror"));
    rpenalty.setToolTipText (parent.translator.lookup ("mirror.tPenalty"));
    rreason.setToolTipText (parent.translator.lookup ("mirror.tReason"));
    rjudge.setToolTipText (parent.translator.lookup ("mirror.tJudge"));
    tableChanged (new TableModelEvent
      (getModel (), TableModelEvent.HEADER_ROW));
    parent.translate (competitor_clear, "mirror.competitor.removeall");
    parent.translate (competitor_add, "mirror.competitor.addall");
    parent.translate (mirror_clear, "mirror.number.clear");
    parent.translate (mirror_random, "mirror.number.random");
    parent.translate (judge_higher, "mirror.judge.higher");
    parent.translate (judge_lower, "mirror.judge.lower");
    parent.translate (judge_add, "mirror.judge.add");
    parent.translate (judge_remove, "mirror.judge.remove");
  }

  /** Invoked when the table model changes */
  public void tableChanged (TableModelEvent e) {
    super.tableChanged (e);
    if (e == null || e.getFirstRow () == TableModelEvent.HEADER_ROW) {
      updateCompetitors ();
      if (rmirror != null && rpenalty != null && rjudge != null) {
	rmirror.setHorizontalAlignment (JLabel.RIGHT);
	rpenalty.setHorizontalAlignment (JLabel.RIGHT);
	rjudge.setHorizontalAlignment (JLabel.RIGHT);
	getColumnModel ().getColumn (COLUMN_MIRROR).setCellRenderer
	  (rmirror);
	getColumnModel ().getColumn (COLUMN_PENALTY).setCellRenderer
	  (rpenalty);
	getTableHeader ().setReorderingAllowed (false);
	getTableHeader ().addMouseListener (headerlistener);
      }
    }
  }

  /** Update the look and feel */
  public void updateUI () {
    super.updateUI ();
    if (judgepopup != null && competitorpopup != null && mirrorpopup != null) {
      SwingUtilities.updateComponentTreeUI (judgepopup);
      SwingUtilities.updateComponentTreeUI (competitorpopup);
      SwingUtilities.updateComponentTreeUI (mirrorpopup);
    }
  }

  /** Update the data on available competitors */
  public void updateCompetitors () {
    if (competitorbox != null) competitorbox.updateCompetitors ();
    frequent = new Hashtable ();
  }

  /** Get a bitmap of too frequent points given by a judge
   * @param judge	the judge
   * @return		a bitmap of too frequent points
   */
  private boolean[] tooFrequent (Judge judge) {
    boolean freq[] = (boolean[]) frequent.get (judge);
    if (freq == null) {
      final Contest c = contest.getSelected ();
      if (c == null)
	frequent = new Hashtable ();
      else
	frequent.put (judge, freq = c.getFrequentPoints (judge));
    }
    return freq;
  }

  /** A combo box for picking a competitor */
  private final class CompetitorBox
    extends JComboBox
  {
    CompetitorBox () {
      updateCompetitors ();
    }
    /** Rebuild the list */
    void updateCompetitors () {
      TreeSet set = new TreeSet (new Comparator ()
	{
	  final Collator collator =
	    Collator.getInstance (parent.getLocale ());
	  public int compare (Object o1, Object o2) {
	    return collator.compare (((Competitor) o1).getName (),
				     ((Competitor) o2).getName ());
	  }
	});
      final Contest c = contest.getSelected ();
      for (Iterator it = parent.series.getCompetitors ().iterator ();
	   it.hasNext (); ) {
	Competitor competitor = (Competitor) it.next ();
	if (c != null && c.getMirror (competitor) == null)
	  set.add (competitor);
      }
      removeAllItems ();
      addItem (MirrorTable.empty);
      for (Iterator i = set.iterator (); i.hasNext (); )
	addItem (i.next ());
    }
  }

  /** A renderer for mirror numbers */
  final class MirrorRenderer extends DefaultTableCellRenderer
  {
    public Component getTableCellRendererComponent (JTable table,
						    Object value,
						    boolean isSelected,
						    boolean hasFocus,
						    int row,
						    int column) {
      final Contest c = MirrorTable.this.contest.getSelected ();
      setBackground (c != null && c.getNumOccurrences ((Integer) value) > 1
		     ? Color.red
		     : null);
      return super.getTableCellRendererComponent
	(table, value, isSelected, hasFocus, row, column);
    }
  }

  /** A renderer for points */
  final class ScoreRenderer extends DefaultTableCellRenderer
  {
    public Component getTableCellRendererComponent (JTable table,
						    Object value,
						    boolean isSelected,
						    boolean hasFocus,
						    int row,
						    int column) {
      final Contest c = MirrorTable.this.contest.getSelected ();
      final Judge judge =
	c != null && column >= COLUMN_JUDGE && c.getNumJudges () > 0
	? c.getJudge (column - COLUMN_JUDGE)
	: null;
      setBackground (judge != null && value != null &&
		     tooFrequent (judge) [((Integer) value).intValue ()]
		     ? Color.red
		     : null);
      return super.getTableCellRendererComponent
	(table, value, isSelected, hasFocus, row, column);
    }
  }

  /** competitor popup menu listener */
  private final class CompetitorListener implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      Object src = e.getSource ();
      final Contest c = contest.getSelected ();
      if (c == null)
	return;
      if (src == competitor_clear) {
	if (parent.confirm ("mirror.competitor.removeall", c.getName ())) {
	  c.removeCompetitors ();
	  tableChanged (null);
	}
      }
      else if (src == competitor_add) {
	if (parent.confirm ("mirror.competitor.addall", c.getName ())) {
	  c.addCompetitors ();
	  tableChanged (null);
	}
      }
    }
  }

  /** mirror popup menu listener */
  private final class MirrorListener implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      Object src = e.getSource ();
      final Contest c = contest.getSelected ();
      if (c == null)
	return;
      if (src == mirror_clear) {
	if (parent.confirm ("mirror.mirror.clear", c.getName ())) {
	  for (int i = c.getNumMirrors (); i-- > 0; )
	    c.getMirror (i).setNumber (null);
	  tableChanged (new TableModelEvent
	    (getModel (), 0, c.getNumMirrors (), COLUMN_MIRROR));
	}
      }
      else if (src == mirror_random) {
	RandomDialog r = new RandomDialog (parent, c.getName ());
	if (r.value != null && r.value.intValue () > 0) {
	  c.shuffleMirrors (r.value.intValue ());
	  tableChanged (null);
	}
      }
    }
  }

  /** judge popup menu listener */
  private final class JudgeListener implements ActionListener
  {
    public void actionPerformed (ActionEvent e)
    {
      Object src = e.getSource ();
      final Contest c = contest.getSelected ();
      Judge judge;
      if (c == null)
	return;
      if (src == judge_higher) {
	if (headercolumn <= COLUMN_JUDGE || c.getNumJudges () == 0)
	  return;
	c.swapJudges (headercolumn - COLUMN_JUDGE,
		      headercolumn - COLUMN_JUDGE - 1);
	tableChanged (null);
      }
      else if (src == judge_lower) {
	if (headercolumn >= COLUMN_JUDGE - 1 + c.getNumJudges ())
	  return;
	c.swapJudges (headercolumn - COLUMN_JUDGE,
		      headercolumn - COLUMN_JUDGE + 1);
	tableChanged (null);
      }
      else if (src == judge_remove) {
	if (c.getNumJudges () == 0)
	  return;
	judge = c.getJudge (headercolumn - COLUMN_JUDGE);
	if (c.hasPoints (judge) &&
	    !parent.confirm ("mirror.judge.remove", judge.getName ()))
	  return;
	c.remove (judge);
	tableChanged (null);
      }
      else {
	JMenuItem item = (JMenuItem) src;
	int i = ((JPopupMenu) item.getParent ()).getComponentIndex (item);
	for (Iterator it = parent.series.getJudges ().iterator ();
	     it.hasNext (); ) {
	  judge = (Judge) it.next ();
	  if (!c.hasJudge (judge) && i-- == 0) {
	    c.add (judge);
	    tableChanged (null);
	    break;
	  }
	}
      }
    }
  }
}
